#pragma once

#include <cstdint>
#include <iterator>

#include <RobotAPI/interface/armem/query.h>

#include <ArmarXCore/core/exceptions/LocalException.h>
#include <ArmarXCore/core/logging/Logging.h>
#include <ArmarXCore/core/exceptions/local/ExpressionException.h>

#include <RobotAPI/libraries/armem/core/error.h>
#include <RobotAPI/libraries/armem/core/ice_conversions.h>

#include "BaseQueryProcessorBase.h"
#include "RobotAPI/libraries/armem/core/Time.h"


namespace armarx::armem::base::query_proc
{

    /**
     * @brief Handles memory queries.
     */
    template <class _EntityT>
    class EntityQueryProcessorBase :
        virtual public BaseQueryProcessorBase<_EntityT, armem::query::data::EntityQuery>
    {
        using Base = BaseQueryProcessorBase<_EntityT, armem::query::data::EntityQuery>;

    public:
        using EntityT = _EntityT;
        using EntitySnapshotT = typename EntityT::EntitySnapshotT;

        using Base::process;
        void process(_EntityT& result,
                     const armem::query::data::EntityQuery& query,
                     const _EntityT& entity) const override
        {
            if (auto q = dynamic_cast<const armem::query::data::entity::All*>(&query))
            {
                process(result, *q, entity);
            }
            else if (auto q = dynamic_cast<const armem::query::data::entity::Single*>(&query))
            {
                process(result, *q, entity);
            }
            else if (auto q = dynamic_cast<const armem::query::data::entity::TimeRange*>(&query))
            {
                process(result, *q, entity);
            }
            else if (auto q = dynamic_cast<const armem::query::data::entity::IndexRange*>(&query))
            {
                process(result, *q, entity);
            }
            else if (auto q = dynamic_cast<const armem::query::data::entity::TimeApprox*>(&query))
            {
                process(result, *q, entity);
            }
            else if (auto q = dynamic_cast<const armem::query::data::entity::BeforeOrAtTime*>(&query))
            {
                process(result, *q, entity);
            }
            else if (auto q = dynamic_cast<const armem::query::data::entity::BeforeTime*>(&query))
            {
                process(result, *q, entity);
            }
            else
            {
                throw armem::error::UnknownQueryType("entity snapshot", query);
            }
        }

        void process(_EntityT& result,
                     const armem::query::data::entity::All& query,
                     const _EntityT& entity) const
        {
            (void) query;
            // Copy this entitiy and its contents.

            for (const auto& [time, snapshot] : entity)
            {
                addResultSnapshot(result, snapshot);
            }
        }

        void process(_EntityT& result,
                     const armem::query::data::entity::Single& query,
                     const _EntityT& entity) const
        {
            if (query.timestamp < 0)
            {
                try
                {
                    addResultSnapshot(result, entity.getLatestSnapshot());
                }
                catch (const armem::error::EntityHistoryEmpty&)
                {
                    // Leave empty.
                }
            }
            else
            {
                Time time = fromIce<Time>(query.timestamp);
                try
                {
                    addResultSnapshot(result, entity.getSnapshot(time));
                }
                catch (const armem::error::MissingEntry&)
                {
                    // Leave empty.
                }
            }
        }

        void process(_EntityT& result,
                     const armem::query::data::entity::TimeRange& query,
                     const _EntityT& entity) const
        {
            if (query.minTimestamp <= query.maxTimestamp || query.minTimestamp < 0 || query.maxTimestamp < 0)
            {
                Time min = fromIce<Time>(query.minTimestamp);
                Time max = fromIce<Time>(query.maxTimestamp);
                process(result, min, max, entity, query);
            }
        }

        void process(_EntityT& result,
                     const armem::query::data::entity::IndexRange& query,
                     const _EntityT& entity) const
        {
            if (entity.empty())
            {
                return;
            }

            size_t first = negativeIndexSemantics(query.first, entity.history().size());
            size_t last = negativeIndexSemantics(query.last, entity.history().size());

            if (first <= last)
            {
                auto it = entity.begin();
                std::advance(it, first);

                size_t num = last - first + 1;  // +1 to make last inclusive
                for (size_t i = 0; i < num; ++i, ++it)
                {
                    addResultSnapshot(result, it);
                }
            }
        }

        void process(_EntityT& result,
                     const Time& min,
                     const Time& max,
                     const _EntityT& entity,
                     const armem::query::data::EntityQuery& query) const
        {
            (void) query;

            // Returns an iterator pointing to the first element that is not less than (i.e. greater or equal to) key.
            auto begin = min.toMicroSeconds() > 0 ? entity.history().lower_bound(min) : entity.history().begin();
            // Returns an iterator pointing to the first element that is *greater than* key.
            auto end = max.toMicroSeconds() > 0 ? entity.history().upper_bound(max) : entity.history().end();

            for (auto it = begin; it != end && it != entity.history().end(); ++it)
            {
                addResultSnapshot(result, it);
            }
        }


        inline auto lastElementBeforeOrAt(const auto& history, const auto timestamp) const
        {
            // first element equal or greater
            typename std::map<Time, EntitySnapshotT>::const_iterator refItFwd = history.upper_bound(timestamp);

            // last element less than
            typename std::map<Time, EntitySnapshotT>::const_iterator refItFwdLt = std::prev(refItFwd);

            // last element not greater than => if this is invalid, we have no suitable elements
            typename std::map<Time, EntitySnapshotT>::const_reverse_iterator refIt(refItFwd);
            if (refIt == history.rend())
            {
                return history.end();
            }

            // now either refItFwd is a perfect match ...
            if (refItFwd->first == timestamp)
            {
                return refItFwd;
            }

            // ... or we return the element before if possible
            if (refItFwdLt != history.begin())
            {
                return refItFwdLt;
            }

            return history.end();
        }

        inline auto lastElementBefore(const auto& history, const auto& timestamp) const
        {
            // first element equal or greater
            typename std::map<Time, EntitySnapshotT>::const_iterator refItFwd = history.upper_bound(timestamp);

            // last element less than
            typename std::map<Time, EntitySnapshotT>::const_iterator refItFwdLt = std::prev(refItFwd);

            // last element not greater than => if this is invalid, we have no suitable elements
            typename std::map<Time, EntitySnapshotT>::const_reverse_iterator refIt(refItFwd);
            if (refIt == history.rend())
            {
                return history.end();
            }

            // we return the element before if possible
            if (refItFwdLt != history.begin())
            {
                return refItFwdLt;
            }

            return history.end();
        }


        void process(_EntityT& result,
                     const armem::query::data::entity::BeforeOrAtTime& query,
                     const _EntityT& entity) const
        {
            const auto referenceTimestamp = fromIce<Time>(query.timestamp);
            ARMARX_CHECK(referenceTimestamp.toMicroSeconds() >= 0) << "Reference timestamp is negative!";

            const auto beforeOrAt = lastElementBeforeOrAt(entity.history(), referenceTimestamp);

            if (beforeOrAt != entity.history().end())
            {
                addResultSnapshot(result, beforeOrAt);
            }
        }


        void process(_EntityT& result,
                     const armem::query::data::entity::BeforeTime& query,
                     const _EntityT& entity) const
        {
            const auto referenceTimestamp = fromIce<Time>(query.timestamp);
            ARMARX_CHECK(referenceTimestamp.toMicroSeconds() >= 0) << "Reference timestamp is negative!";

            const auto maxEntries = fromIce<std::int64_t>(query.maxEntries);

            const auto itBefore = lastElementBefore(entity.history(), referenceTimestamp);
            if (itBefore == entity.history().end())
            {
                ARMARX_WARNING << "No suitable element found";
                return;
            }

            const auto nEntriesBefore = std::distance(entity.history().begin(), itBefore);

            const int nEntries = [&]()
            {
                // see query.ice
                if (maxEntries > 0)
                {
                    return std::min(nEntriesBefore, maxEntries);
                }

                return nEntriesBefore; // all elements before timestamp
            }
            ();

            auto it = itBefore;
            for (std::int64_t i = 0; i < nEntries; i++, --it)
            {
                addResultSnapshot(result, it);
            }
        }

        void process(_EntityT& result,
                     const armem::query::data::entity::TimeApprox& query,
                     const _EntityT& entity) const
        {

            const auto referenceTimestamp = fromIce<Time>(query.timestamp);
            ARMARX_CHECK(referenceTimestamp.toMicroSeconds() >= 0) << "Reference timestamp is negative!";

            const auto referenceTimestampMicroSeconds = referenceTimestamp.toMicroSeconds();
            const auto epsDuration = fromIce<Time>(query.eps).toMicroSeconds();

            // elements have to be in range [t_ref - eps, t_ref + eps] if eps is positive
            const auto isInRange = [&](const Time & t) -> bool
            {
                if (epsDuration <= 0)
                {
                    return true;
                }

                return std::abs(t.toMicroSeconds() - referenceTimestampMicroSeconds) <= epsDuration;
            };

            // first element before or at timestamp
            const auto beforeOrAt = lastElementBeforeOrAt(entity.history(), referenceTimestamp);
            const bool isBeforeOrAtValid = beforeOrAt != entity.history().end();
            const auto isPerfectMatch = isBeforeOrAtValid && beforeOrAt->first == referenceTimestamp;

            // first element greater
            const auto after = entity.history().upper_bound(referenceTimestamp);
            const bool isAfterValid = after != entity.history().end();

            // if both are invalid, there is nothing to be gained here.
            if ((not isBeforeOrAtValid) and (not isAfterValid))
            {
                const armem::Time dt = referenceTimestamp - entity.getLatestTimestamp();

                ARMARX_WARNING << "Lookup " << dt.toMilliSeconds() << " ms into the future.";
                return;
            } 
            // -> now one or both are valid ...

            // is 'before' a perfect match?
            if(isPerfectMatch)
            {
                addResultSnapshot(result, beforeOrAt);
                return;
            }

            // only 'before' valid?
            if (not isAfterValid)
            {
                if (isInRange(beforeOrAt->first))
                {
                    addResultSnapshot(result, beforeOrAt);
                }
                return;
            }

            // only 'after' valid?
            if (not isBeforeOrAtValid)
            {
                if (isInRange(after->first))
                {
                    addResultSnapshot(result, after);
                }
                return;
            }
            // -> now both are valid

            // return both (if in range) => user can interpolate
            if (isInRange(beforeOrAt->first))
            {
                addResultSnapshot(result, beforeOrAt);
            }

            if (isInRange(after->first))
            {
                addResultSnapshot(result, after);
            }
            
        }


        static size_t negativeIndexSemantics(long index, size_t size)
        {
            const size_t max = size > 0 ? size - 1 : 0;
            if (index >= 0)
            {
                return std::clamp<size_t>(static_cast<size_t>(index), 0, max);
            }
            else
            {
                return static_cast<size_t>(std::clamp<long>(static_cast<long>(size) + index, 0, static_cast<long>(max)));
            }
        }

    protected:
        virtual void addResultSnapshot(_EntityT& result, typename _EntityT::ContainerT::const_iterator it) const = 0;
        virtual void addResultSnapshot(_EntityT& result, const typename _EntityT::EntitySnapshotT& snapshot) const = 0;

    };
}
